import socket

from ckan.lib import mailer

# TODO: put this in the config/interface so that it can be overridden
# TODO: add html version of the body
default_save_body = """
Hello,

As requested, a DOI has successfully been created for your search.
The DOI will become available at https://doi.org/{}, though this can sometimes take a few hours.

Please ensure that you cite this DOI whenever you use this data! Follow the DOI link for more
details.

Best wishes,
The NHM Data Portal Bot
""".strip()


def send_saved_search_email(email_address, doi):
    # send the DOI to the user in an email
    try:
        mailer.mail_recipient(
            recipient_email=email_address,
            recipient_name='DOI Requester',
            subject='Query DOI created',
            body=default_save_body.format(doi.doi),
        )
        return True
    except (mailer.MailerException, socket.error):
        # the error will be logged automatically by CKAN's mailing functions
        return False



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK

import logging
import random
import string
from datetime import datetime
from typing import Optional, Tuple

from ckan import model
from ckan.common import asbool
from ckan.plugins import toolkit
from datacite import DataCiteMDSClient, schema41
from datacite.errors import DataCiteError, DataCiteNotFoundError

from ckanext.query_dois.lib.query import Query
from ckanext.query_dois.model import QueryDOI

log = logging.getLogger(__name__)

# this is a test prefix available for all minters to use for testing purposes
TEST_PREFIX = '10.5072'


def is_test_mode():
    """
    Should we use the test datacite MDS API?

    :returns: True if we should, False if not. Defaults to True.
    """
    return asbool(toolkit.config.get('ckanext.query_dois.test_mode', True))


def get_prefix():
    """
    Gets the prefix to use for the DOIs we mint.

    :returns: the prefix to use for the new DOIs
    """
    prefix = toolkit.config.get('ckanext.query_dois.prefix')

    if prefix is None:
        raise TypeError('You must set the ckanext.query_dois.prefix config value')

    if prefix == TEST_PREFIX:
        raise ValueError(
            f'The test prefix {TEST_PREFIX} has been retired, use a prefix defined in '
            f'your datacite test account'
        )

    return prefix


def get_client():
    """
    Get a datacite MDS API client, configured for use.

    :returns: a DataCite client object
    """
    kwargs = dict(
        username=toolkit.config.get('ckanext.query_dois.datacite_username'),
        password=toolkit.config.get('ckanext.query_dois.datacite_password'),
        prefix=get_prefix(),
        test_mode=is_test_mode(),
    )
    # datacite 1.0.1 isn't updated for the test prefix deprecation yet so this is a temp fix
    if is_test_mode():
        kwargs.update({'url': 'https://mds.test.datacite.org'})
    return DataCiteMDSClient(**kwargs)


def generate_doi(client):
    """
    Generate a new DOI which isn't currently in use. The database is checked for
    previous usage, as is Datacite itself. Use whatever value is retuned from this
    function quickly to avoid double use as this function uses no locking.

    :param client: an instance of the DataCiteMDSClient class
    :returns: the full, unique DOI
    """
    # the list of valid characters is larger than just lowercase and the digits but we don't need
    # that many options and URLs with just alphanumeric characters in them are nicer. We just use
    # lowercase characters to avoid any issues with case being ignored
    valid_characters = string.ascii_lowercase + string.digits

    attempts = 5

    while attempts > 0:
        # generate a random 8 character identifier and prepend qd. to make it easier to identify
        # DOIs from this extension
        identifier = f'qd.{"".join(random.choice(valid_characters) for _ in range(8))}'
        # form the doi using the prefix
        doi = f'{get_prefix()}/{identifier}'

        # check this doi doesn't exist in the table
        if model.Session.query(QueryDOI).filter(QueryDOI.doi == doi).count():
            continue

        # check against the datacite service
        try:
            client.metadata_get(doi)
            # if a doi is found, we need to try again
            continue
        except DataCiteNotFoundError:
            # if no doi is found, we're good!
            pass
        except DataCiteError as e:
            log.warning(
                f'Error whilst checking new DOIs with DataCite. DOI: {doi}, error: {e}'
            )
            attempts -= 1
            continue

        # if we've made it this far the doi isn't in the database and it's not in datacite already
        return doi
    else:
        raise Exception('Failed to generate a DOI')


def find_existing_doi(query: Query) -> Optional[QueryDOI]:
    """
    Returns a QueryDOI object representing the query, or returns None if one doesn't
    exist.

    :param query: a Query object
    :returns: a QueryDOI object or None
    """
    return (
        model.Session.query(QueryDOI)
        .filter(
            QueryDOI.query_hash == query.query_hash,
            QueryDOI.query_version == query.query_version,
            QueryDOI.resources_and_versions == query.resources_and_versions,
        )
        .first()
    )


def create_doi_on_datacite(
    client: DataCiteMDSClient, doi: str, timestamp: datetime, query: Query
):
    """
    Mints the given DOI on datacite using the client.

    :param client: the MDS datacite client
    :param doi: the doi (full, prefix and suffix)
    :param timestamp: the datetime when the DOI was created
    :param query: a Query object
    """
    # create the data for datacite
    data = {
        'identifier': {
            'identifier': doi,
            'identifierType': 'DOI',
        },
        'creators': [{'creatorName': author} for author in query.authors],
        'titles': [
            {
                'title': toolkit.config.get('ckanext.query_dois.doi_title').format(
                    count=query.count
                )
            }
        ],
        'publisher': toolkit.config.get('ckanext.query_dois.publisher'),
        'publicationYear': str(timestamp.year),
        'resourceType': {'resourceTypeGeneral': 'Dataset'},
    }

    # use an assert here because the data should be valid every time, otherwise it's something the
    # developer is going to have to fix
    assert schema41.validate(data)

    # create the metadata on datacite
    client.metadata_post(schema41.tostring(data))

    # create the URL the DOI will point to, i.e. the landing page
    data_centre, identifier = doi.split('/')
    landing_page_url = toolkit.url_for(
        'query_doi.landing_page', data_centre=data_centre, identifier=identifier
    )
    site = toolkit.config.get('ckan.site_url')
    if site[-1] == '/':
        site = site[:-1]
    # mint the DOI
    client.doi_post(doi, site + landing_page_url)


def create_database_entry(
    doi: str,
    query: Query,
    timestamp: datetime,
):
    """
    Inserts the database row for the query DOI.

    :param doi: the doi (full, prefix and suffix)
    :param query: the query
    :param timestamp: the datetime the DOI was created
    :returns: the QueryDOI object
    """
    query_doi = QueryDOI(
        doi=doi,
        timestamp=timestamp,
        resources_and_versions=query.resources_and_versions,
        requested_version=query.version,
        query=query.query,
        query_version=query.query_version,
        query_hash=query.query_hash,
        count=query.count,
        resource_counts=query.counts,
    )
    query_doi.save()
    return query_doi


def mint_multisearch_doi(query: Query) -> Tuple[bool, QueryDOI]:
    """
    Mint a DOI on datacite using their API and create a new QueryDOI object, saving it
    to the database. If we already have a query which would produce identical data to
    the one passed then we return the existing QueryDOI object and don't mint or insert
    anything.

    This function handles DOIs created for the versioned datastore's multisearch action.

    :param query: the query
    :returns: a boolean indicating whether a new DOI was minted and the QueryDOI object
        representing the query's DOI
    """
    # check if there are any dois already for this query
    existing_doi = find_existing_doi(query)
    if existing_doi is not None:
        return False, existing_doi

    # generate a new DOI to store this query against
    timestamp = datetime.now()
    client = get_client()
    doi = generate_doi(client)
    create_doi_on_datacite(client, doi, timestamp, query)
    query_doi = create_database_entry(doi, query, timestamp)
    return True, query_doi



import pytest
from ckan import model

from ckanext.query_dois.model import query_doi_stat_table, query_doi_table


@pytest.fixture
def setup_db():
    for table in (query_doi_table, query_doi_stat_table):
        if not table.exists(model.meta.engine):
            table.create(model.meta.engine)



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK

import copy
import itertools
import json
import operator
from collections import OrderedDict
from dataclasses import asdict, dataclass
from typing import Optional
from urllib.parse import urlencode

from ckan import model
from ckan.plugins import toolkit

from ..lib.stats import DOWNLOAD_ACTION, SAVE_ACTION
from ..lib.utils import get_resource_and_package
from ..model import QueryDOI, QueryDOIStat

column_param_mapping = (
    ('doi', QueryDOIStat.doi),
    ('identifier', QueryDOIStat.identifier),
    ('domain', QueryDOIStat.domain),
    ('action', QueryDOIStat.action),
)


@dataclass
class InaccessibleResource:
    id: Optional[str] = None
    name: Optional[str] = None
    package_id: Optional[str] = None
    package_name: Optional[str] = None
    package_title: Optional[str] = None
    record_count: int = 0

    @property
    def is_unknown(self):
        return self.name is None

    def as_dict(self):
        return asdict(self)


def get_query_doi(doi):
    """
    Retrieves a QueryDOI object from the database for the given DOI, if there is one,
    otherwise returns None.

    :param doi: the doi (full doi, prefix/suffix)
    :returns: A QueryDOI object or None
    """
    return model.Session.query(QueryDOI).filter(QueryDOI.doi == doi).first()


def get_authors(packages):
    """
    Retrieves all the authors from the given packages, de-duplicates them (if necessary)
    and then returns them as a list.

    Note that this function takes a list of packages as it is multi-package and
    therefore multi-resource ready.

    :param packages: the packages
    :returns: a list of author(s)
    """
    # use an ordered dict in the absence of a sorted set
    authors = OrderedDict()
    for package in packages:
        author = package['author']
        # some author values will contain many authors with a separator, perhaps , or ;
        for separator in (';', ','):
            if separator in author:
                authors.update({a: True for a in author.split(separator)})
                break
        else:
            # if the author value didn't contain a separator then we can just use the value as is
            authors[author] = True

    return list(authors.keys())


def encode_params(params, version=None, extras=None, for_api=False):
    """
    Encodes the parameters for a query in the CKAK resource view format and returns as a
    query string.

    :param params: a dict of parameters, such as a DatastoreQuery's query dict
    :param version: the version to add into the query string (default: None)
    :param extras: an optional dict of extra parameters to add as well as the ones found
        in the params dict (default: None)
    :param for_api: whether the query string is for a CKAN resource view or an API get
        as it changes the format (default: False)
    :returns: a query string of the query parameters (no ? at the start but will include
        & if needed)
    """
    query_string = {}
    extras = [] if extras is None else extras.items()
    # build the query string from the dicts we have first
    for param, value in itertools.chain(params.items(), extras):
        # make sure to ignore all version data in the dicts
        if param == 'version':
            continue
        if param == 'filters':
            value = copy.deepcopy(value)
            if version is None:
                value.pop('__version__', None)
        query_string[param] = value

    # now add the version in if needed
    if version is not None:
        query_string.setdefault('filters', {})['__version__'] = version

    # finally format any nested dicts correctly (this is for the filters field basically)
    for param, value in query_string.items():
        if isinstance(value, dict):
            if for_api:
                # the API takes the data in JSON format so we just need to serialise it
                value = json.dumps(value)
            else:
                # if the data is going in a query string for a resource view it needs to be
                # encoded in a special way
                parts = []
                for sub_key, sub_value in value.items():
                    if not isinstance(sub_value, list):
                        sub_value = [sub_value]
                    parts.extend('{}:{}'.format(sub_key, v) for v in sub_value)
                value = '|'.join(parts)
            query_string[param] = value

    return urlencode(query_string)


def generate_rerun_urls(resource, package, query, rounded_version=None):
    """
    Generate a dict containing all the "rerun" URLs needed to allow the user to revisit the data
    either through the website or through the API. The dict returned will look like following:

        {
            "page": {
                "original": ...
                "current": ...
            },
            "api": {
                "original": ...
                "current": ...
            }
        }

    :param resource: the resource dict
    :param package: the package dict
    :param query: the query dict
    :param rounded_version: the version rounded down to the nearest available on the resource
    :returns: a dict of urls
    """
    page_url = toolkit.url_for(
        'resource.read', id=package['name'], resource_id=resource['id']
    )
    api_url = '/api/action/datastore_search'
    api_extras = {'resource_id': resource['id']}
    url_dict = {
        'page': {
            'current': page_url + '?' + encode_params(query),
        }
    }
    if rounded_version is not None:
        url_dict['page']['original'] = (
            page_url + '?' + encode_params(query, version=rounded_version)
        )
        url_dict['api'] = {
            'current': api_url
            + '?'
            + encode_params(query, extras=api_extras, for_api=True),
            'original': api_url
            + '?'
            + encode_params(
                query, version=rounded_version, extras=api_extras, for_api=True
            ),
        }
    return url_dict


def get_stats(query_doi):
    """
    Retrieve some simple stats about the query DOI - this includes the total downloads and the
    last download timestamp. Note that we are specifically looking for downloads here, no other
    actions are considered.

    :param query_doi: the QueryDOI object
    :returns: a 3-tuple containing the total downloads, total saves and the last download timestamp
    """
    # count how many download stats we have on this doi
    download_total = (
        model.Session.query(QueryDOIStat)
        .filter(QueryDOIStat.doi == query_doi.doi)
        .filter(QueryDOIStat.action == DOWNLOAD_ACTION)
        .count()
    )
    # count how many save stats we have on this doi
    save_total = (
        model.Session.query(QueryDOIStat)
        .filter(QueryDOIStat.doi == query_doi.doi)
        .filter(QueryDOIStat.action == SAVE_ACTION)
        .count()
    )
    # find the last stats object we have for this doi
    last = (
        model.Session.query(QueryDOIStat)
        .filter(QueryDOIStat.doi == query_doi.doi)
        .filter(QueryDOIStat.action == DOWNLOAD_ACTION)
        .order_by(QueryDOIStat.id.desc())
        .first()
    )
    return download_total, save_total, last.timestamp if last is not None else None


def render_datastore_search_doi_page(query_doi):
    """
    Renders a DOI landing page for a datastore_search based query DOI.

    :param query_doi: the query DOI
    :returns: the rendered page
    """
    # currently we only deal with single resource query DOIs
    resource_id = query_doi.get_resource_ids()[0]
    rounded_version = query_doi.get_rounded_versions()[0]

    try:
        resource, package = get_resource_and_package(resource_id)
        is_inaccessible = False
        in_datastore = resource.get('datastore_active', False)
    except (toolkit.ObjectNotFound, toolkit.NotAuthorized):
        resource = None
        package = None
        in_datastore = False
        is_inaccessible = True

    # we ignore the saves count as it will always be 0 for a datastore_search DOI
    downloads, _saves, last_download_timestamp = get_stats(query_doi)
    usage_stats = {
        'downloads': downloads,
        'last_download_timestamp': last_download_timestamp,
    }

    # warnings
    warnings = []
    if is_inaccessible:
        warnings = [
            toolkit._(
                'All resources associated with this search have been deleted, moved, '
                'or are no longer available in their previous format.'
            )
        ]
    elif not in_datastore:
        warnings = [
            toolkit._(
                'All records associated with this search have been removed from the '
                'search index. The data may still exist, but they are no longer '
                'versioned and cannot be filtered.'
            )
        ]

    context = {
        'query_doi': query_doi,
        'doi': query_doi.doi,
        'resource': resource,
        'package': package,
        'version': rounded_version,
        'usage_stats': usage_stats,
        'is_inaccessible': is_inaccessible,
        'in_datastore': in_datastore,
        'warnings': warnings,
        # these are defaults for if the resource is inaccessible
        'package_doi': None,
        'authors': toolkit._('Unknown'),
        'reruns': {},
    }

    if not is_inaccessible:
        context.update(
            {
                # this is effectively an integration point with the ckanext-doi
                # extension. If there is demand we should open this up so that we can
                # support other dois on packages extensions
                'package_doi': (
                    package['doi'] if package.get('doi_status', False) else None
                ),
                'authors': get_authors([package]),
                'reruns': generate_rerun_urls(
                    resource,
                    package,
                    query_doi.query,
                    rounded_version if in_datastore else None,
                ),
            }
        )

    return toolkit.render('query_dois/single_landing_page.html', context)


def create_current_slug(query_doi: QueryDOI, ignore_resources=None) -> str:
    """
    Creates a slug for the given query DOI at the current version, this is done with a
    nav slug which has no version.

    :param query_doi: the QueryDOI
    :param ignore_resources: a list of resource IDs to ignore
    :returns: a slug
    """
    resource_ids = query_doi.get_resource_ids()
    if ignore_resources:
        resource_ids = [r for r in resource_ids if r not in ignore_resources]

    slug_data_dict = {
        'query': query_doi.query,
        'query_version': query_doi.query_version,
        'resource_ids': resource_ids,
        'nav_slug': True,
    }
    current_slug = toolkit.get_action('vds_slug_create')({}, slug_data_dict)
    return current_slug['slug']


def render_multisearch_doi_page(query_doi: QueryDOI):
    """
    Renders a DOI landing page for a datastore_multisearch based query DOI.

    :param query_doi: the query DOI
    :returns: the rendered page
    """
    resource_ids = set(query_doi.get_resource_ids())

    accessible_packages = {}
    accessible_resources = {}
    inaccessible_resources = []

    offset = 0
    action = toolkit.get_action('current_package_list_with_resources')
    unchecked_resource_ids = resource_ids.copy()
    while len(unchecked_resource_ids) > 0:
        # DOI resources should always be public, so ignore anything that isn't
        context = {'ignore_auth': False, 'user': None}
        packages = action(context, {'offset': offset, 'limit': 200})
        if not packages:
            break
        for package in packages:
            package_accessible_resources = []
            for resource in package.get('resources', []):
                if resource['id'] in resource_ids:
                    unchecked_resource_ids.discard(resource['id'])
                    if resource['datastore_active']:
                        package_accessible_resources.append(resource['id'])
                        accessible_resources[resource['id']] = {
                            'name': resource['name'],
                            'package_id': package['id'],
                        }
                    else:
                        # non-datastore resources are "inaccessible" as DOI resources,
                        # but we can still show their details
                        inaccessible_resources.append(
                            InaccessibleResource(
                                id=resource['id'],
                                name=resource['name'],
                                package_id=package['id'],
                                package_name=package['name'],
                                package_title=package['title'],
                            )
                        )
            if package_accessible_resources:
                accessible_packages[package['id']] = {
                    'title': package['title'],
                    'name': package['name'],
                    'resource_ids': package_accessible_resources,
                }
        offset += len(packages)

    # if it couldn't be found, we don't know what it is
    for rid in unchecked_resource_ids:
        inaccessible_resources.append(InaccessibleResource(id=rid))

    inaccessible_count = len(inaccessible_resources)

    # usage stats
    downloads, saves, last_download_timestamp = get_stats(query_doi)
    usage_stats = {
        'downloads': downloads,
        'saves': saves,
        'last_download_timestamp': last_download_timestamp,
    }

    # current details
    sorted_resource_counts = sorted(
        [
            (k, v)
            for k, v in query_doi.resource_counts.items()
            if k in accessible_resources
        ],
        key=operator.itemgetter(1),
        reverse=True,
    )
    current_details = {
        'resource_count': len(accessible_resources),
        'package_count': len(accessible_packages),
        'sorted_resource_counts': sorted_resource_counts,
        'record_count': query_doi.count
        if inaccessible_count == 0
        else sum([v for k, v in sorted_resource_counts]),
    }

    # saved details
    if inaccessible_count == 0:
        saved_details = {
            'resource_count': len(accessible_resources),
            'record_count': query_doi.count,
            'missing_resources': 0,
            'missing_records': 0,
        }
    else:
        saved_details = {
            'resource_count': len(query_doi.resource_counts),
            'record_count': query_doi.count,
            'missing_resources': inaccessible_count,
            'missing_records': query_doi.count - current_details['record_count'],
        }

    # warnings
    warnings = []
    if len(accessible_resources) == 0:
        current_slug = None
        warnings = [
            toolkit._(
                'All resources associated with this search have been deleted, moved, '
                'or are no longer available in their previous format.'
            )
        ]
    else:
        current_slug = create_current_slug(
            query_doi, ignore_resources=[r.id for r in inaccessible_resources]
        )
        if inaccessible_count > 0:
            warnings.append(
                toolkit._(
                    'Some resources have been deleted, moved, or are no longer '
                    'available. Affected resources: '
                )
                + str(inaccessible_count)
            )

    # inaccessible resources
    unknown = {'resource_count': 0, 'record_count': 0}
    known = []
    for res in inaccessible_resources:
        if res.is_unknown:
            unknown['resource_count'] += 1
            unknown['record_count'] += query_doi.resource_counts[res.id]
        else:
            res.record_count = query_doi.resource_counts[res.id]
            known.append(res.as_dict())
    inaccessible_resource_details = known
    if unknown['resource_count'] > 0:
        inaccessible_resource_details.append(
            InaccessibleResource(
                id=None,
                name=' '.join(
                    [toolkit._('Unknown resources'), f'({unknown["resource_count"]})']
                ),
                package_title=toolkit._('Unknown package'),
                record_count=unknown['record_count'],
            ).as_dict()
        )

    context = {
        'query_doi': query_doi,
        'original_slug': query_doi.doi,
        'current_slug': current_slug,
        'usage_stats': usage_stats,
        'resources': accessible_resources,
        'packages': accessible_packages,
        'details': current_details,
        'saved_details': saved_details,
        'has_changed': inaccessible_count > 0,
        'is_inaccessible': len(accessible_resources) == 0,
        'warnings': warnings,
        'inaccessible_resources': inaccessible_resource_details,
    }
    return toolkit.render('query_dois/multisearch_landing_page.html', context)



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-query-dois
# Created by the Natural History Museum in London, UK

from ckan.plugins import toolkit


def get_resource_and_package(resource_id):
    """
    Given a resource ID, returns the resource's dict and the parent package's dict too.

    :param resource_id: the resource ID
    :returns: a 2-tuple, containing the resource dict and the package dict
    """
    resource = toolkit.get_action('resource_show')({}, {'id': resource_id})
    package = toolkit.get_action('package_show')({}, {'id': resource['package_id']})
    return resource, package



import random
import string
import time
from datetime import datetime
from unittest.mock import MagicMock
from uuid import uuid4

import pytest
from ckan.tests import factories

from ckanext.query_dois.helpers import get_doi_count, get_most_recent_dois
from ckanext.query_dois.lib.doi import create_database_entry
from ckanext.query_dois.model import QueryDOI


def make_doi(resource_id) -> QueryDOI:
    random_str = ''.join(random.choice(string.ascii_lowercase) for _ in range(10))
    version = int(time.time() * 1000)
    return create_database_entry(
        doi=random_str,
        query=MagicMock(
            resources_and_versions={resource_id: version},
            version=version,
            query={},
            query_version='v1.0.0',
            query_hash=str(uuid4()),
            count=4,
            counts={resource_id: 4},
        ),
        timestamp=datetime.now(),
    )


@pytest.mark.usefixtures('clean_db', 'setup_db')
class TestGetDOICount:
    def test_when_dois_have_been_made(self):
        # create 2 packages
        package_1 = factories.Dataset(name='package1')
        package_2 = factories.Dataset(name='package2')
        # create 2 resources per package
        resource_1 = factories.Resource(package_id=package_1['id'])
        resource_2 = factories.Resource(package_id=package_1['id'])
        resource_3 = factories.Resource(package_id=package_2['id'])
        resource_4 = factories.Resource(package_id=package_2['id'])

        dois = [
            make_doi(resource_1['id']),
            make_doi(resource_1['id']),
            make_doi(resource_2['id']),
            make_doi(resource_2['id']),
            make_doi(resource_2['id']),
            make_doi(resource_3['id']),
            make_doi(resource_4['id']),
            make_doi(resource_4['id']),
            make_doi(resource_4['id']),
            make_doi(resource_4['id']),
            make_doi(resource_4['id']),
            make_doi(resource_4['id']),
        ]
        assert get_doi_count(package_1['id']) == 5
        assert get_doi_count(package_2['id']) == 7

    def test_no_dois_have_been_made(self):
        # create 2 packages
        package_1 = factories.Dataset(name='package1')
        package_2 = factories.Dataset(name='package2')
        # create 2 resources per package
        resource_1 = factories.Resource(package_id=package_1['id'])
        resource_2 = factories.Resource(package_id=package_1['id'])
        resource_3 = factories.Resource(package_id=package_2['id'])
        resource_4 = factories.Resource(package_id=package_2['id'])
        # create a doi on a resource in package 2
        make_doi(resource_3['id'])
        assert get_doi_count(package_1['id']) == 0

    def test_no_resources(self):
        # create a package
        package_1 = factories.Dataset(name='package1')
        # create no resources
        assert get_doi_count(package_1['id']) == 0

    def test_package_not_exist(self):
        assert get_doi_count('fiowenfwuibefuiwbefuiwbef') == 0


@pytest.mark.usefixtures('clean_db', 'setup_db')
class TestGetMostRecentDOIs:
    def test_no_package(self):
        assert len(get_most_recent_dois('efiownfwe', 5)) == 0

    def test_no_resources(self):
        package_1 = factories.Dataset(name='package1')
        assert len(get_most_recent_dois(package_1['id'], 5)) == 0

    def test_no_dois(self):
        package_1 = factories.Dataset(name='package1')
        resource_1 = factories.Resource(package_id=package_1['id'])
        assert len(get_most_recent_dois(package_1['id'], 5)) == 0

    def test_dois_less_than_limit(self):
        package_1 = factories.Dataset(name='package1')
        resource_1 = factories.Resource(package_id=package_1['id'])
        resource_2 = factories.Resource(package_id=package_1['id'])
        make_doi(resource_1['id'])
        make_doi(resource_2['id'])
        assert len(get_most_recent_dois(package_1['id'], 5)) == 2

    def test_dois_more_than_limit(self):
        package_1 = factories.Dataset(name='package1')
        resource_1 = factories.Resource(package_id=package_1['id'])
        for _ in range(10):
            make_doi(resource_1['id'])
        assert len(get_most_recent_dois(package_1['id'], 5)) == 5



import pytest
from ckan.tests import factories
from ckan.tests.helpers import call_action

from ckanext.query_dois.lib.utils import get_resource_and_package


@pytest.mark.usefixtures('clean_db')
@pytest.mark.filterwarnings('ignore::sqlalchemy.exc.SADeprecationWarning')
def test_get_resource_and_package():
    package = factories.Dataset()
    resource = factories.Resource(package_id=package['id'])

    shown_resource = call_action('resource_show', id=resource['id'])
    shown_package = call_action('package_show', id=package['id'])

    assert get_resource_and_package(resource['id']) == (shown_resource, shown_package)



"""
Set old query version.

Revision ID: a74242a670e0
Revises:
Create Date: 2023-06-09 12:20:05.632095
"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declarative_base

# revision identifiers, used by Alembic.
revision = 'a74242a670e0'
down_revision = None
branch_labels = None
depends_on = None

Base = declarative_base()


class QueryDOI(Base):
    __tablename__ = 'query_doi'
    id = sa.Column(sa.UnicodeText, primary_key=True)
    query_version = sa.Column(sa.UnicodeText, nullable=True)


def upgrade():
    bind = op.get_bind()
    session = orm.Session(bind=bind)

    for query_doi in session.query(QueryDOI).filter(QueryDOI.query_version.is_(None)):
        query_doi.query_version = 'v0'

    session.commit()


def downgrade():
    bind = op.get_bind()
    session = orm.Session(bind=bind)

    for query_doi in session.query(QueryDOI).filter(QueryDOI.query_version == 'v0'):
        query_doi.query_version = None

    session.commit()
